import requests


class RestClient:
    def __init__(self, tesseract_pro):
        self.api_version = 'v1'
        self._tpro = tesseract_pro

    def get(self, endpoint, params=None):
        url = f"{self._tpro.server_url}/{self.api_version}/api/{endpoint}"
        headers = {"x-tesseractpro-api-key": self._tpro.api_token}

        response = requests.get(url, headers=headers, params=params)

        return response.json()

    def post(self, endpoint, data=None):
        url = f"{self._tpro.server_url}/{self.api_version}/api/{endpoint}"
        headers = {"x-tesseractpro-api-key": self._tpro.api_token}
        response = requests.post(url, headers=headers, json=data)

        data = response.json()

        if isinstance(data, dict):
            if data['statusCode'] and data['statusCode'] == 400:
                raise Exception(data['message'])

        return data



from .utils.event_emitter import EventEmitter
from .timewindow import TimeWindow


class Chart(EventEmitter):
    ON_TOOL_UPDATE = "tool:update"
    ON_TOOL_ADD = "tool:add"
    ON_TOOL_REMOVE = "tool:remove"
    ON_ALERT_TRIGGER = "alert:trigger"
    ON_ALERT_ADD = "alert:add"
    ON_ALERT_UPDATE = "alert:update"
    ON_ALERT_REMOVE = "alert:remove"
    ON_PRICE = "price"

    def __init__(self,
                 tesseract_pro,
                 space_id: str,
                 timeframe: int,
                 symbol: str,
                 time_window: TimeWindow
                 ):
        super().__init__()

        self.space_id = space_id
        self.timeframe = timeframe
        self.time = time_window
        self.symbol = symbol
        self.tpro = tesseract_pro

        self.tpro.market.subscribe(symbol)

        # create event listeners
        def create_handler(context, event):
            return lambda data: self.handle_event(context, event, data)

        self.tpro.tool.on("update", create_handler("tool", "update"))
        self.tpro.tool.on("add", create_handler("tool", "add"))
        self.tpro.tool.on("remove", create_handler("tool", "remove"))

        self.tpro.alert.on("trigger", create_handler("alert", "trigger"))
        self.tpro.alert.on("add", create_handler("alert", "add"))
        self.tpro.alert.on("update", create_handler("alert", "update"))
        self.tpro.alert.on("remove", create_handler("alert", "remove"))

        self.tpro.market.on(
            "pricechange", lambda data: self.handle_price_change(data))

    def handle_price_change(self, data):
        if data[0] == self.symbol:
            self.emit('price', data[1])

    def handle_event(self, context, event, data):
        # Event may not be bound to timeframe
        if "timeframe" not in data:
            self.emit(f'tool:{event}', data)
            return

        # Not for us
        if data['timeframe'] != self.timeframe:
            return

        # Lets go!
        self.emit(f'{context}:{event}', data)

    def get_candles(self):
        data = self.tpro.rest.post('ohlc', {
            "from": self.time.get_start_time(),
            "count": self.time.candles,
            "timeframe": self.timeframe,
            "market": self.symbol
        })

        return data

    def get_tools(self):
        # candles = self.time.get_start_time() - self.time.get_end_time()
        # print(candles / (self.timeframe * 60))
        # print(self.time.get_start_time(), self.time.get_end_time())

        data = self.tpro.rest.post('tools', {
            "from": self.time.get_start_time(),
            "to": self.time.get_end_time(),
            "timeframe": self.timeframe * 60,
            "market": self.symbol,
            "space_id": self.space_id
        })

        return data

    def wait_for_events(self):
        self.tpro.wait_for_events()



from .base import Base
from .decorators.on import on


class Tools (Base):
    @on("tool:add")
    def on_tools_added(self, data) -> None:
        self.emit('add', data)

    @on("tool:update")
    def on_tool_update(self, data) -> None:
        self.emit('update', data)

    @on("tool:remove")
    def on_tool_remove(self, data) -> None:
        self.emit('remove', data)

    def get_tools(
            self,
            from_unix: int,
            to_unix: int,
            timeframe: int,
            market: str,
            space_id: str
    ):
        """
        Return tools for the given parameters
        """
        data = self._tpro.rest.post("tools", {
            "to": to_unix,
            "from": from_unix,
            "timeframe": timeframe * 60,
            "market": market,
            "space_id": space_id
        })
        return {item['id']: item for item in data}



from typing import Union
from .decorators.on import on
from .base import Base


class Markets (Base):
    def subscribe(self, symbol: Union[tuple[str, ...], str]) -> None:
        """
        Subscribe to one or more markets events
        """

        if isinstance(symbol, str):
            symbol = [symbol]

        self._tpro._socket.emit(
            'market:subscribe',
            {'symbols': symbol},
            namespace="/api"
        )

    @on("market:pricechange")
    def on_price_change(self, data: tuple[str, int]) -> None:
        self.emit('pricechange', (data[0], data[1]))



from .base import Base
from .decorators.on import on


class Alerts (Base):
    @on("alert:trigger")
    def on_alert_trigger(self, data) -> None:
        self.emit('trigger', data)

    @on("alert:add")
    def on_tools_added(self, data) -> None:
        self.emit('add', data)

    @on("alert:update")
    def on_tool_update(self, data) -> None:
        self.emit('update', data)

    @on("alert:remove")
    def on_tool_remove(self, data) -> None:
        self.emit('remove', data)




class TimeWindow:
    def __init__(self, timeframe: int, start_time: int, candles: int):
        self.timeframe = timeframe
        self.time = start_time
        self.candles = candles

    def get_start_time(self) -> int:
        return self.time

    def get_end_time(self) -> int:
        return self.time - (self.candles * self.timeframe * 60)
